#
# LandLand 1 + 2 Decorruptor
# Copyright (c) GreaseMonkey, 2021.
# 
# Permission to use, copy, modify, and/or distribute this software for any purpose with or without fee is hereby granted.
# 
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#

from __future__ import annotations

import argparse
import struct

# Some notes...

# A block is 128 bytes.
# 0x06FB -> 0x077B Shirkagawa (+1 block)
# Dupe: 0x8480
# 0x15B3 -> 0x1633 Back Room (+1 block)
# Dupe: 0x9200
# 0x06ED -> 0x076D Secret Cavern (+1 block)
# Dupe: 0xA680
# 0x05CC -> 0x06CC Secret Cavern East x1 (+2 blocks)
# Dupe: 0xAB80
# Dupe: 0xAF00
# 0x150B -> 0x---- Secret Cavern East x2


def main() -> None:
    parser = argparse.ArgumentParser(description="Decorrupts Land Land 1 and 2.")
    parser.add_argument("infname", help="Input file")
    parser.add_argument("--outfname", "-o", required=True, help="Output file")
    args = parser.parse_args()

    infname: str
    outfname: str
    infname = args.infname
    outfname = args.outfname

    decorrupt(infname=infname, outfname=outfname)
    size = validate(fname=outfname)
    truncate(fname=outfname, size=size)
    resize = validate(fname=outfname)
    assert size == resize


def decorrupt(*, infname: str, outfname: str) -> None:
    """Decorrupts a world."""

    with open(infname, "rb") as infp:
        with open(outfname, "wb") as outfp:
            prev_block = b""
            tsub = 0
            while True:
                t = infp.tell()
                next_block = infp.read(128)
                if next_block == b"":
                    break
                if next_block == prev_block and t >= 0x00200:
                    tsub += 0x80
                    print(f"Dupe at 0x{t:05X} -> 0x{t-tsub:05X}")
                else:
                    outfp.write(next_block)
                prev_block = next_block

    print("-"*76)
    print("")


def validate(*, fname: str) -> int:
    """Validates a world, returning the length."""

    with open(fname, "rb") as fp:
        # Read header for board count
        board_count: int
        _, board_count, = struct.unpack("<HH", fp.read(4))
        fp.seek(0x200)
        board_count += 1

        print(f"Board count: {board_count}")
        for board_idx in range(board_count):
            this_t: int = fp.tell()
            board_len: int
            board_len, = struct.unpack("<H", fp.read(2))
            next_t: int = this_t + 2 + board_len

            # Read title
            title_len: int = fp.read(1)[0]
            title_untrimmed: bytes = fp.read(50)
            title_trimmed: bytes = title_untrimmed[:title_len]
            print(f"Title: {title_len:3d} {title_trimmed!r:54s} {title_untrimmed!r}")
            if title_len > 50:
                raise Exception(f"unusually long title length {title_len:d}")

            # Load RLE
            rle_count = 0
            rle_accum = 0
            for i in range(60*25):
                if rle_count == 0:
                    rle_count = fp.read(3)[0]
                    rle_accum += rle_count
                    #print(f"RLE {i:4d} {rle_count:3d} {rle_accum:4d}")
                rle_count = (rle_count - 1) & 0xFF

            if rle_count != 0:
                raise Exception(f"RLE count nonzero at end (got {rle_count:d})")

            # Skip most of header 2, but get the stat count
            fp.read(86)
            stat_count: int
            stat_count, = struct.unpack("<H", fp.read(2))
            stat_count += 1

            # Read stats
            print(f"Stat count: {stat_count}")
            for stat_idx in range(stat_count):
                # Skip most of the stat info
                fp.read(23)
                # Read the length
                code_len: int
                code_len, = struct.unpack("<h", fp.read(2))
                # Skip the extra 8 unused bytes
                fp.read(8)
                if code_len > 0:
                    # Skip the code
                    fp.read(code_len)

            got_t = fp.tell()
            if next_t != got_t:
                raise Exception(f"offset mismatch: Expected 0x{next_t:05X}, got 0x{got_t:05X} instead")

        print(f"End: 0x{fp.tell():05X}")
        result = fp.tell()

    print("-"*76)
    print("")

    return result


def truncate(*, fname: str, size: int) -> None:
    """Truncates a world to a given size."""

    # POSIX has a truncate() syscall, but this should be portable.
    with open(fname, "rb") as infp:
        data = infp.read()
    data = data[:size]
    assert len(data) == size, f"Size mismatch: got 0x{len(data):05X}, expected 0x{size:05X}"
    with open(fname, "wb") as outfp:
        outfp.write(data)


if __name__ == "__main__":
    main()
